# STL
import re
from abc import ABC, abstractmethod
from typing import Set, List, Type
from functools import lru_cache as cache  # cache comes in 3.9

# PDM
import regex
from typing_extensions import override

# LOCAL
from sonatoki.utils import prep_dictionary
from sonatoki.constants import (
    VOWELS,
    NIMI_PU,
    ALPHABET,
    ALL_PUNCT,
    ALLOWABLES,
    CONSONANTS,
    IGNORABLES,
    NIMI_UCSUR,
    NIMI_KU_LILI,
    NIMI_KU_SULI,
    NIMI_LINKU_CORE,
    ALL_PUNCT_RANGES,
    NIMI_PU_SYNONYMS,
    NIMI_LINKU_COMMON,
    NIMI_LINKU_OBSCURE,
    NIMI_LINKU_SANDBOX,
    UCSUR_PUNCT_RANGES,
    NIMI_LINKU_UNCOMMON,
)

regex.DEFAULT_VERSION = regex.VERSION1


class Filter(ABC):
    @classmethod
    @abstractmethod
    @cache(maxsize=None)
    def filter(cls, token: str) -> bool:
        raise NotImplementedError


class MinLen(Filter):
    """
    Meta filter meant to be inherited by another filter to add a length requirement.
    Multiple-inherit with `MinLen` as the first argument so `super()` resolves correctly.
    You may also construct any other filter with a minimum length filter like so:

    ```
    MinLen(Alphabetic, 3)
    ```
    """

    length = 0

    @classmethod
    @cache(maxsize=None)
    def filter(cls, token: str) -> bool:
        if len(token) < cls.length:
            return False
        return super().filter(token)

    def __new__(cls, filter: Type[Filter], length_: int) -> Type[Filter]:
        class MinLenFilter(MinLen, Filter):
            length = length_

        return MinLenFilter


class RegexFilter(Filter):
    pattern: "re.Pattern[str]"

    @classmethod
    @override
    @cache(maxsize=None)
    def filter(cls, token: str) -> bool:
        return not not re.fullmatch(cls.pattern, token)


class Regex1Filter(Filter):
    pattern: "regex.Pattern[str]"

    @classmethod
    @override
    @cache(maxsize=None)
    def filter(cls, token: str) -> bool:
        return not not regex.fullmatch(cls.pattern, token)


class MemberFilter(Filter):
    tokens: Set[str]

    @classmethod
    @override
    @cache(maxsize=None)
    def filter(cls, token: str) -> bool:
        return token.lower() in cls.tokens


class SubsetFilter(Filter):
    tokens: Set[str]

    @classmethod
    @override
    @cache(maxsize=None)
    def filter(cls, token: str) -> bool:
        return set(token.lower()).issubset(cls.tokens)


class Miscellaneous(MemberFilter):
    tokens = prep_dictionary(ALLOWABLES)


class EnglishIgnorables(MemberFilter):
    """NOTE: Not recommended for use.
    It is better to use a Long* filter such as LongSyllabic than to use this filter.
    This filter hides words from scoring rather than scoring them poorly,
    which is more of a benefit than a loss for a word you would like to omit."""

    tokens = prep_dictionary(IGNORABLES)


class ProperName(Filter):
    """Determines if a given token is a valid name (also called a loan word).
    When Toki Pona is written with the Latin alphabet, names are generally
    capitalized at their start. This filter identifies those tokens.

    Note that this alone cannot determine if a token is a valid name,
    because a standalone name is considered invalid in Toki Pona- names
    generally have head nouns. This tool only examines one token at a
    time, so cannot detect names any better than identifying their
    capital letter.
    """

    @classmethod
    @override
    @cache(maxsize=None)
    def filter(cls, token: str) -> bool:
        return token == token.capitalize()
        # TODO:  If the token is in a script which doesn't have a case distinction,
        # this will errantly match.


class LongProperName(MinLen, ProperName):
    length = 2  # reject "names" of length 1


class NimiPu(MemberFilter):
    tokens = prep_dictionary(NIMI_PU)


class NimiPuSynonyms(MemberFilter):
    tokens = prep_dictionary(NIMI_PU_SYNONYMS)


class NimiKuSuli(MemberFilter):
    tokens = prep_dictionary(NIMI_KU_SULI)


class NimiKuLili(MemberFilter):
    tokens = prep_dictionary(NIMI_KU_LILI)


class NimiLinkuCore(MemberFilter):
    tokens = prep_dictionary(NIMI_LINKU_CORE)


class NimiLinkuCommon(MemberFilter):
    tokens = prep_dictionary(NIMI_LINKU_COMMON)


class NimiLinkuUncommon(MemberFilter):
    tokens = prep_dictionary(NIMI_LINKU_UNCOMMON)


class NimiLinkuObscure(MemberFilter):
    tokens = prep_dictionary(NIMI_LINKU_OBSCURE)


class NimiLinkuSandbox(MemberFilter):
    tokens = prep_dictionary(NIMI_LINKU_SANDBOX)


class NimiUCSUR(MemberFilter):
    tokens = prep_dictionary(NIMI_UCSUR)


class Phonotactic(RegexFilter):
    """Determines if a given token is phonotactically valid Toki Pona (or `n`).

    Excludes both consecutive nasals and the illegal syllables:
    - "nm", "nn"
    - "wu", "wo", "ji", "ti"

    Note that if this validator is used after `Cleaners.ConsecutiveDuplicates`,
    "nn" cannot be found.
    """

    pattern = re.compile(
        rf"^((^[{VOWELS}]|[klmnps][{VOWELS}]|[jt][aeou]|[w][aei])(n(?![mn]))?)+$|^n$",
        # Can't split initial vowel group off like in Syllabics because of
        # consecutive nasal detection; it is costly to duplicate
        flags=re.IGNORECASE,
    )


class LongPhonotactic(MinLen, Phonotactic):
    length = 3


class Syllabic(RegexFilter):
    """Determines if a given token is syllabically valid Toki Pona (or `n`).

    Words must have correctly ordered vowels and consonants, but the
    phonotactic exceptions are not considered.
    """

    # rf"^((^[{VOWELS}]|[{CONSONANTS}][{VOWELS}])n?)+$|^n$"
    # Alterative I was exploring takes ~15% more steps
    pattern = re.compile(
        rf"^(?:^[{VOWELS}]n?)?(?:[{CONSONANTS}][{VOWELS}]n?)*$|^n$",
        flags=re.IGNORECASE,
    )


class LongSyllabic(MinLen, Syllabic):
    length = 3


class Alphabetic(SubsetFilter):
    tokens = set(ALPHABET)


class AlphabeticRe(RegexFilter):
    pattern = re.compile(rf"[{ALPHABET}]+", flags=re.IGNORECASE)


class LongAlphabetic(MinLen, Alphabetic):
    length = 3


class Numeric(Filter):
    """Determine if a given token is entirely numeric. Covers all numeric
    symbols in Unicode.

    This will fail to find numeric tokens such as "1.111" or "-42",
    but if used with the aggressive tokenizer designed for `tok`, these will be
    split into `["1", ".", "111"]` and `["-", "42"]` respectively. As such, the
    numeric tokens will be split from their punctuation.
    """

    @classmethod
    @override
    @cache(maxsize=None)
    def filter(cls, msg: str) -> bool:
        return msg.isnumeric()


class Punctuation(SubsetFilter):
    """Identify whether a token is entirely punctuation.

    Fastest implementation.
    """

    tokens = set(ALL_PUNCT)


class PunctuationRe(RegexFilter):
    """Faster implementation of `PunctuationRe1`.

    Goes out of date compared to the `regex` library if UNICODE_PUNCT_RANGES is not updated.
    """

    pattern = re.compile(rf"[{ALL_PUNCT_RANGES}]+")


class PunctuationRe1(Regex1Filter):
    """Reference implementation for identifying tokens made entirely of
    punctuation."""

    pattern = regex.compile(
        rf"[\p{{Punctuation}}\p{{posix_punct}}{UCSUR_PUNCT_RANGES}]+"
    )


class OrFilter:
    """Instantiate with more than one filter to compose them into one filter,
    returning True when any individual filter matches or False otherwise.
    Requires at least two filters.

    OrFilter exists as a compromise between the need to score some
    filters equally, while not adding custom behavior to scorers. I
    could have allowed a position to have a list of filters instead of
    one filter, but this would require cleaning the user's input, and
    nested handling of lists. It also would not have been as powerful- I
    would need another param for the and/or switch, or to not give users
    the choice.

    Instead, the user is responsible for building an OrFilter out of
    their desired filters.
    """

    @staticmethod
    def __generic_filter(*filters_: Type[Filter]) -> Type[Filter]:

        class CombinedFilter(Filter):
            filters: List[Type[Filter]] = list(filters_)  # TODO: tuple better?

            @classmethod
            @override
            @cache(maxsize=None)
            def filter(cls, token: str) -> bool:
                for f in cls.filters:
                    if f.filter(token):
                        return True
                return False

        return CombinedFilter

    def __new__(cls, *filters: Type[Filter]) -> Type[Filter]:
        if not len(filters) >= 2:
            raise ValueError("Provide at least two Filters to OrFilter.")

        member_filters = [f for f in filters if issubclass(f, MemberFilter)]
        if len(member_filters) >= 2:
            raise Warning("Use OrMemberFilter for combining two or more MemberFilters.")

        filter = cls.__generic_filter(*filters)

        return filter


class OrMemberFilter:
    @staticmethod
    def __member_filter(*filters: Type[MemberFilter]) -> Type[MemberFilter]:
        all_token_sets: List[Set[str]] = [f.tokens for f in filters]
        all_tokens: Set[str] = set().union(*all_token_sets)

        class CombinedFilter(MemberFilter):
            tokens = all_tokens

        return CombinedFilter

    def __new__(cls, *filters_: Type[MemberFilter]) -> Type[MemberFilter]:
        if not len(filters_) >= 2:
            raise ValueError("Provide two or more MemberFilters to OrMemberFilter.")
        filter = cls.__member_filter(*filters_)
        return filter


class AndFilter:
    """Instantiate with more than one filter to compose them into one filter,
    returning False when any individual filter fails to match or True
    otherwise.

    Requires at least two filters.
    """

    def __new__(cls, *filters_: Type[Filter]) -> Type[Filter]:
        if not len(filters_) >= 2:
            raise ValueError("Must provide at least two Filters to AndFilter.")

        class AnonymousAndFilter(Filter):
            filters: List[Type[Filter]] = list(filters_)  # TODO: tuple better?

            @classmethod
            @override
            @cache(maxsize=None)
            def filter(cls, token: str) -> bool:
                for f in cls.filters:
                    if not f.filter(token):
                        return False
                return True

        return AnonymousAndFilter


__all__ = [
    "Alphabetic",
    "AndFilter",
    "EnglishIgnorables",
    "LongAlphabetic",
    "LongPhonotactic",
    "LongProperName",
    "LongSyllabic",
    "MinLen",
    "NimiLinkuCore",
    "NimiLinkuSandbox",
    "NimiPu",
    "NimiPuSynonyms",
    "NimiUCSUR",
    "Numeric",
    "OrFilter",
    "Phonotactic",
    "ProperName",
    "Punctuation",
    "Syllabic",
]



# STL
import re
import itertools
from typing import Set, List, TypeVar, Iterable

# LOCAL
from sonatoki.Cleaners import Lowercase, ConsecutiveDuplicates

TO_ESCAPE = ["\\", "^", "[", "]", "-"]

T = TypeVar("T")


def prep_dictionary(words: Iterable[str]) -> Set[str]:
    out: Set[str] = set()
    cleaners = [Lowercase, ConsecutiveDuplicates]
    for word in words:
        for c in cleaners:
            word = c.clean(word)
        out.add(word)
    return out


def regex_escape(s: str) -> str:
    """Escape all characters which must be escaped when embedded in a character
    class."""
    for c in TO_ESCAPE:
        s = s.replace(c, rf"\{c}")  # one backslash
    return s


def to_range(start: int, prev: int) -> str:
    if start == prev:
        return rf"\U{start:08x}"
    return rf"\U{start:08x}-\U{prev:08x}"


def find_unicode_ranges(chars: str) -> List[str]:
    if not chars:
        return []

    s_chars = sorted(set(chars))

    ranges: List[str] = []
    start = ord(s_chars[0])
    prev = start

    for i in range(1, len(s_chars)):
        cur = ord(s_chars[i])
        if cur == prev + 1:  # range is still contiguous
            prev = cur
            continue

        ranges.append(to_range(start, prev))
        start = prev = cur

    last = ord(s_chars[-1])
    ranges.append(to_range(start, last))

    return ranges


def find_unicode_chars(ranges: List[str]) -> str:
    result: List[str] = []
    for item in ranges:
        if "-" in item:
            start, end = item.split("-")
            start = int(start.lstrip("\\U"), 16)
            end = int(end.lstrip("\\U"), 16)
            result.extend(chr(code_point) for code_point in range(start, end + 1))
        else:
            result.append(chr(int(item.lstrip("\\U"), 16)))
    return "".join(result)


def overlapping_pairs(iterable: Iterable[T]) -> Iterable[T]:
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    return overlapping_ntuples(iterable, n=2)


def overlapping_ntuples(iterable: Iterable[T], n: int) -> Iterable[T]:
    teed = itertools.tee(iterable, n)
    for i in range(1, n):
        for j in range(i):
            _ = next(teed[i], None)
            # offset start by position

    # ends when any iter is empty; all groups will be same size
    return zip(*teed)


if __name__ == "__main__":
    """Helper script to fetch UNICODE_PUNCT in constants.py."""

    PUNCT_CATEGORIES = {
        "Pc",
        "Pd",
        "Pe",
        "Pf",
        "Pi",
        "Po",
        "Ps",
        "Sm",
        "Sk",
        "Sc",
        "So",
    }
    # Connector, Dash, Close (end), Final, Initial, Other, Open (sOpen), Math, Modifier (kModifier), Currency, Other

    # NOTE: UnicodeData.txt lists character ranges if there would be many characters.
    # (e.g. CJK Ideograph, First at 4E00 and CJK Ideograph, Last at 9FFF).
    # This does not apply to any currently defined punctuation category.

    EXCEPTION_RANGES = re.compile(r"""[Ⓐ-ⓩ🄰-🅉🅐-🅩🅰-🆉]+""")
    # These groups are in Symbol other (So) but are not part of `\p{Punctuation}`
    # NOTE: There are many characters which look like writing characters but are not. Examples:
    # - kangxi radicals from ⺀ to ⿕ which are for demonstration
    # - circled katakana from  to ㋾ which... shouldn't be in \p{Punctuation} but oh well

    def is_punctuation(data: List[str]):
        return data[2] in PUNCT_CATEGORIES

    def get_character(data: List[str]):
        return chr(int(data[0], 16))

    def is_exception(c: str):
        return not not re.fullmatch(EXCEPTION_RANGES, c)

    # http://www.unicode.org/Public/UNIDATA/UnicodeData.txt
    unicode_punctuation = ""
    with open("UnicodeData.txt", "r") as f:
        for line in f:
            data = line.split(";")
            if not is_punctuation(data):
                continue

            char = get_character(data)
            if is_exception(char):
                continue

            unicode_punctuation += char

    with open("UnicodePunctuation.txt", "w") as f:
        _ = f.write(unicode_punctuation)



# PDM
import hypothesis.strategies as st
from hypothesis import given

# LOCAL
from sonatoki.Filters import (
    NimiPu,
    Syllabic,
    Alphabetic,
    NimiKuLili,
    NimiKuSuli,
    Phonotactic,
    NimiLinkuCore,
    NimiPuSynonyms,
    NimiLinkuCommon,
    NimiLinkuObscure,
    NimiLinkuSandbox,
    NimiLinkuUncommon,
)
from sonatoki.Cleaners import Lowercase, ConsecutiveDuplicates
from sonatoki.constants import (
    NIMI_PU,
    NIMI_KU_LILI,
    NIMI_KU_SULI,
    NIMI_LINKU_CORE,
    NIMI_PU_SYNONYMS,
    NIMI_LINKU_COMMON,
    NIMI_LINKU_OBSCURE,
    NIMI_LINKU_SANDBOX,
    NIMI_LINKU_UNCOMMON,
)


@given(st.sampled_from(list(NIMI_PU | NIMI_PU_SYNONYMS)))
def test_pu_filters_non_overlap(s: str):
    res_pu = NimiPu.filter(s)
    res_synonyms = NimiPuSynonyms.filter(s)
    assert (res_pu + res_synonyms) == 1


@given(st.sampled_from(list(NIMI_KU_SULI | NIMI_KU_LILI)))
def test_ku_filters_non_overlap(s: str):
    res_ku_suli = NimiKuSuli.filter(s)
    res_ku_lili = NimiKuLili.filter(s)
    assert (res_ku_suli + res_ku_lili) == 1


@given(
    st.sampled_from(
        list(
            NIMI_LINKU_CORE
            | NIMI_LINKU_COMMON
            | NIMI_LINKU_UNCOMMON
            | NIMI_LINKU_OBSCURE
            | NIMI_LINKU_SANDBOX
        )
    )
)
def test_linku_filters_non_overlap(s: str):
    s = Lowercase.clean(s)
    s = ConsecutiveDuplicates.clean(s)

    res_core = NimiLinkuCore.filter(s)
    res_common = NimiLinkuCommon.filter(s)
    res_uncommon = NimiLinkuUncommon.filter(s)
    res_obscure = NimiLinkuObscure.filter(s)
    res_sandbox = NimiLinkuSandbox.filter(s)

    assert (res_core + res_common + res_uncommon + res_obscure + res_sandbox) == 1


@given(st.sampled_from(list(NIMI_LINKU_CORE | NIMI_LINKU_COMMON | NIMI_LINKU_UNCOMMON)))
def test_nimi_linku_properties(s: str):
    assert ConsecutiveDuplicates.clean(s) == s, repr(s)
    assert Alphabetic.filter(s), repr(s)
    assert Syllabic.filter(s), repr(s)
    assert Phonotactic.filter(s), repr(s)
    # Passing phonotactic implies all of the above



# PDM
import hypothesis.strategies as st
from hypothesis import given, example

# LOCAL
from sonatoki.Preprocessors import (
    URLs,
    Spoilers,
    AllQuotes,
    Backticks,
    Reference,
    ArrowQuote,
    DoubleQuotes,
    SingleQuotes,
    DiscordEmotes,
    DiscordSpecial,
    DiscordChannels,
    DiscordMentions,
    AngleBracketObject,
)


@given(st.from_regex(URLs.pattern.pattern, fullmatch=True))
@example("https://google.com")
@example("https://mun.la")
@example("https://discord.gg/")
@example("http://example.com")
@example("http://localhost:80")
def test_URLs(s: str):
    assert URLs.process(s).strip() == ""


@given(st.from_regex(Spoilers.pattern.pattern, fullmatch=True))
@example("|| | ||")
@example("|| content\n\n\ncontent ||")
@example("||\n||")
@example("|| \n|\n\n || || ||")
def test_Spoilers(s: str):
    res = Spoilers.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(st.from_regex(Backticks.pattern.pattern, fullmatch=True))
@example("` ` ` `")
def test_Backticks(s: str):
    res = Backticks.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(st.from_regex(ArrowQuote.pattern.pattern, fullmatch=True))
@example("> base")
@example("> newline\n> newline")
def test_ArrowQuote(s: str):
    res = ArrowQuote.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(st.from_regex(DoubleQuotes.pattern.pattern, fullmatch=True))
@example('" "" "')
@example('" "\n" "')
@example('" \n "')
def test_DoubleQuotes(s: str):
    res = DoubleQuotes.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(st.from_regex(SingleQuotes.pattern.pattern, fullmatch=True))
@example("' '' '")
@example("' '\n' '")
@example("' \n '")
def test_SingleQuotes(s: str):
    res = SingleQuotes.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(st.from_regex(DiscordEmotes.pattern.pattern, fullmatch=True))
@example("<a:example:123123>")
@example("<:example:123123>")
def test_DiscordEmotes(s: str):
    res = DiscordEmotes.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(st.from_regex(DiscordMentions.pattern.pattern, fullmatch=True))
@example("<@497549183847497739>")
@example("<@!457890000>")
@example("<@&18398198981985>")
def test_DiscordMentions(s: str):
    res = DiscordMentions.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(st.from_regex(DiscordChannels.pattern.pattern, fullmatch=True))
@example("<#19858915>")
@example("<#18591912589812985>")
def test_DiscordChannels(s: str):
    res = DiscordChannels.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(st.from_regex(DiscordSpecial.pattern.pattern, fullmatch=True))
@example("<id:guide>")
@example("<id:browse>")
def test_DiscordSpecial(s: str):
    res = DiscordSpecial.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(
    st.from_regex(DiscordEmotes.pattern.pattern, fullmatch=True)
    | st.from_regex(DiscordMentions.pattern.pattern, fullmatch=True)
    | st.from_regex(DiscordChannels.pattern.pattern, fullmatch=True)
    | st.from_regex(DiscordSpecial.pattern.pattern, fullmatch=True)
    | st.from_regex(AngleBracketObject.pattern.pattern, fullmatch=True)
)
@example("<https://example.com>")
@example("<#123124125125>")
def test_AngleBracketObject(s: str):
    res = AngleBracketObject.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(
    st.from_regex(SingleQuotes.pattern.pattern, fullmatch=True)
    | st.from_regex(DoubleQuotes.pattern.pattern, fullmatch=True)
    | st.from_regex(Backticks.pattern.pattern, fullmatch=True)
    | st.from_regex(ArrowQuote.pattern.pattern, fullmatch=True)
    | st.from_regex(AllQuotes.pattern.pattern, fullmatch=True)
)
@example("> bruh")
@example("`bruh`")
def test_AllQuotes(s: str):
    res = AllQuotes.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(st.from_regex(Reference.pattern.pattern, fullmatch=True))
@example("[[Brainstorm]]")
@example("[[Phatic Phrases]]")
@example("[[Yahoo!]]")
def test_Reference(s: str):
    res = Reference.process(s).strip()
    assert res == "", (repr(s), repr(res))



# STL
from copy import deepcopy
from typing import List, Type, TypedDict

# PDM
from typing_extensions import NotRequired

# LOCAL
from sonatoki.Filters import (
    Filter,
    Numeric,
    Syllabic,
    NimiUCSUR,
    Alphabetic,
    NimiKuLili,
    NimiKuSuli,
    ProperName,
    Punctuation,
    LongSyllabic,
    Miscellaneous,
    NimiLinkuCore,
    LongAlphabetic,
    LongProperName,
    OrMemberFilter,
    NimiLinkuCommon,
    NimiLinkuObscure,
    NimiLinkuSandbox,
    NimiLinkuUncommon,
)
from sonatoki.Scorers import Number, Scorer, PassFail, SoftScaling, SoftPassFail
from sonatoki.Cleaners import Cleaner, ConsecutiveDuplicates
from sonatoki.Tokenizers import Tokenizer
from sonatoki.Preprocessors import (
    URLs,
    Backticks,
    Reference,
    Preprocessor,
    AngleBracketObject,
)


class IloConfig(TypedDict):
    preprocessors: List[Type[Preprocessor]]
    cleaners: List[Type[Cleaner]]
    ignoring_filters: List[Type[Filter]]
    scoring_filters: List[Type[Filter]]
    scorer: Type[Scorer]
    passing_score: Number
    word_tokenizer: NotRequired[Type[Tokenizer]]
    sent_tokenizer: NotRequired[Type[Tokenizer]]


# TODO: branching configs? config builder?

BaseConfig: IloConfig = {
    "preprocessors": [URLs],
    "cleaners": [ConsecutiveDuplicates],
    "ignoring_filters": [Numeric, Punctuation],
    "scoring_filters": [],
    "scorer": PassFail,
    "passing_score": 0.8,
}


PrefConfig: IloConfig = {
    "preprocessors": [Backticks, URLs, Reference],
    "cleaners": [ConsecutiveDuplicates],
    "ignoring_filters": [Numeric, Punctuation],
    "scoring_filters": [
        OrMemberFilter(NimiLinkuCore, NimiLinkuCommon, NimiUCSUR, Miscellaneous),
        LongSyllabic,
        LongProperName,
        LongAlphabetic,
    ],
    "scorer": SoftScaling,
    "passing_score": 0.8,
}

CorpusConfig: IloConfig = {
    "preprocessors": [Backticks, URLs, AngleBracketObject, Reference],
    "cleaners": [ConsecutiveDuplicates],
    "ignoring_filters": [Numeric, Punctuation],
    "scoring_filters": [
        OrMemberFilter(
            NimiLinkuCore,
            NimiLinkuCommon,
            NimiLinkuUncommon,
            NimiLinkuObscure,
            NimiLinkuSandbox,
            NimiUCSUR,
            Miscellaneous,
        ),
        LongSyllabic,
        LongProperName,
        LongAlphabetic,
    ],
    "scorer": SoftScaling,
    "passing_score": 0.8,
}
"""Mimics the previous implementation of ilo pi toki pona taso."""
LazyConfig: IloConfig = {
    "preprocessors": [Backticks, URLs, AngleBracketObject, Reference],
    "cleaners": [ConsecutiveDuplicates],
    "ignoring_filters": [Numeric, Punctuation],
    "scoring_filters": [Alphabetic, NimiUCSUR, ProperName, Miscellaneous],
    "scorer": SoftPassFail,
    "passing_score": 0.8,
}
"""This is extremely silly."""
IsipinEpikuConfig: IloConfig = {
    "preprocessors": [Backticks, URLs, AngleBracketObject, Reference],
    "cleaners": [ConsecutiveDuplicates],
    "ignoring_filters": [Numeric, Punctuation],
    "scoring_filters": [
        OrMemberFilter(
            NimiKuSuli,
            NimiKuLili,
            NimiLinkuUncommon,
            NimiLinkuObscure,
            NimiLinkuSandbox,
        ),
        LongSyllabic,
        LongProperName,
        LongAlphabetic,
    ],
    "scorer": SoftScaling,
    "passing_score": 0.8,
}


DiscordConfig: IloConfig = {
    "preprocessors": [Backticks, URLs, AngleBracketObject, Reference],
    "cleaners": [ConsecutiveDuplicates],
    "ignoring_filters": [Numeric, Punctuation],
    "scoring_filters": [
        OrMemberFilter(NimiLinkuCore, NimiLinkuCommon, NimiUCSUR, Miscellaneous),
        LongSyllabic,
        LongProperName,
        LongAlphabetic,
    ],
    "scorer": SoftScaling,
    "passing_score": 0.8,
}

TelegramConfig: IloConfig = deepcopy(PrefConfig)
ForumConfig: IloConfig = deepcopy(PrefConfig)


__all__ = [
    "BaseConfig",
    "CorpusConfig",
    "DiscordConfig",
    "ForumConfig",
    "IloConfig",
    "LazyConfig",
    "PrefConfig",
    "TelegramConfig",
]



# STL
import math
from abc import ABC, abstractmethod
from typing import Dict, List, Type, Union

# PDM
from typing_extensions import override

# LOCAL
from sonatoki.Filters import Filter

Number = Union[int, float]
Weights = Dict[str, Number]


class Scorer(ABC):
    @classmethod
    @abstractmethod
    def score(cls, tokens: List[str], filters: List[Type[Filter]]) -> Number:
        """Score a list of tokens using the given `Filter`s, returning a
        `Number` between 0 and 1 inclusive."""
        raise NotImplementedError


class Soften(Scorer):
    """Meta `Scorer` which scales the scores of short messages to reduce the
    impact of shortness on scoring.

    The scores of short messages are scaled by mapping the token count
    to [0.5, 1.0] via the sigmoid function, then raising the score to
    the resultant power.

    For example, a single token scoring 0.64 will score 0.8 instead.
    """

    @staticmethod
    def sigmoid(n: int) -> Number:
        return 1 / (1 + math.exp(-(0.30 * (n - 1))))
        # n-1 makes sigmoid(1) == 0.5
        # 0.30 softens scaling in favor of short input
        # return n / (1+abs(n))   # too weak in 0.7+

    @classmethod
    @override
    def score(cls, tokens: List[str], filters: List[Type[Filter]]) -> Number:
        percentage = super().score(tokens, filters)  # type: ignore [abstractmethod]
        len_tokens = len(tokens)
        percentage **= cls.sigmoid(len_tokens)
        return percentage

    def __new__(cls, scorer: Type[Scorer]) -> Type[Scorer]:
        class SoftenedScorer(Soften, scorer): ...

        return SoftenedScorer


class PassFail(Scorer):
    """If a token matches any filter, it scores 1.

    Otherwise, it scores 0.
    """

    @classmethod
    def score_token(cls, token: str, filters: List[Type[Filter]]) -> Number:
        for f in filters:
            if f.filter(token):
                return 1
        return 0

    @classmethod
    @override
    def score(cls, tokens: List[str], filters: List[Type[Filter]]) -> Number:
        if not tokens:
            return 1

        total_score = 0
        len_tokens = len(tokens)
        for token in tokens:
            total_score += cls.score_token(token, filters)
        return total_score / len_tokens if len_tokens else 0


class Scaling(Scorer):
    """Tokens score 1 for matching the first filter, and a linearly reduced
    amount for matching later filters based on how many filters there are.

    For example, if there are 4 filters, a token scores 1.0, 0.75, 0.50,
    and 0.25 for matching each respectively.

    In other words, filter order matters, weighing earlier listed
    filters higher than later ones. This is desirable to avoid messages
    which would only match weaker filters, as these are less likely to
    be Toki Pona.
    """

    @classmethod
    def score_token(cls, token: str, filters: List[Type[Filter]], scale: int):
        for i, f in enumerate(filters):
            if f.filter(token):
                return scale - i
        return 0

    @classmethod
    @override
    def score(cls, tokens: List[str], filters: List[Type[Filter]]) -> Number:
        if not tokens:
            return 1

        total_score = 0
        len_filters = len(filters)
        max_score = len(tokens) * len_filters
        for token in tokens:
            total_score += cls.score_token(token, filters, len_filters)
        return total_score / max_score if max_score else 0


class SoftPassFail(Soften, PassFail):
    """Same as `PassFail`, but shorter messages are subject to less harsh
    scoring."""


class SoftScaling(Soften, Scaling):
    """Same as `Scaling`, but shorter messages are subject to less harsh
    scoring."""


# class Logarithmic(Scorer): ...


__all__ = ["PassFail", "SoftPassFail", "Scaling", "SoftScaling"]



# STL

# PDM
import hypothesis.strategies as st
from hypothesis import given, assume, example

# LOCAL
from sonatoki.utils import overlapping_pairs
from sonatoki.Cleaners import Lowercase, ConsecutiveDuplicates, ConsecutiveDuplicatesRe

# FILESYSTEM
from .test_utils import PROPER_NAME_RE


@given(st.from_regex(ConsecutiveDuplicatesRe.pattern.pattern))
@example("tooooki a")
@example("muuuuuu")
@example("nnn")
@example("")
@example("manna")  # syllabically but not phonotactically valid
def test_ConsecutiveDuplicatesRe(s: str):
    _ = assume("\n" not in s)
    res = ConsecutiveDuplicatesRe.clean(s)
    for a, b in overlapping_pairs(res):
        assert a.lower() != b.lower(), (s, res)


@given(st.from_regex(ConsecutiveDuplicatesRe.pattern.pattern))
@example("Aaa")
@example("aAa")
@example("aaA")
@example("BbbbrrrRRRUUuuuhHhHhH")
def test_ConsecutiveDuplicates(s: str):
    _ = assume("\n" not in s)
    res_re = ConsecutiveDuplicatesRe.clean(s)
    res_fn = ConsecutiveDuplicates.clean(s)
    assert res_re == res_fn, repr(s)


@given(st.from_regex(PROPER_NAME_RE))
def test_Lowercase(s: str):
    cleaned = Lowercase.clean(s)
    assert cleaned == s.lower()  # yeah really



# STL
import re
from abc import ABC, abstractmethod
from typing import Set, List

# PDM
import regex
from typing_extensions import override, deprecated

# LOCAL
from sonatoki.utils import regex_escape
from sonatoki.Filters import NimiUCSUR  # seriously this sucks
from sonatoki.constants import ALL_PUNCT, SENTENCE_PUNCT, ALL_PUNCT_RANGES

regex.DEFAULT_VERSION = regex.VERSION1


class Tokenizer(ABC):
    @classmethod
    @abstractmethod
    def tokenize(cls, s: str) -> List[str]: ...


class SetTokenizer(Tokenizer):
    delimiters: Set[str]


class RegexTokenizer(Tokenizer):
    pattern: "re.Pattern[str]"

    @classmethod
    @override
    def tokenize(cls, s: str) -> List[str]:
        return [clean for word in re.split(cls.pattern, s) if (clean := word.strip())]


class Regex1Tokenizer(Tokenizer):
    pattern: "regex.Pattern[str]"

    @classmethod
    @override
    def tokenize(cls, s: str) -> List[str]:
        return [
            clean for word in regex.split(cls.pattern, s) if (clean := word.strip())
        ]


class WordTokenizer(SetTokenizer):
    delimiters = set(ALL_PUNCT)

    @classmethod
    def __helper(cls, s: str, tokens: List[str], last_match: int, i: int):
        match = s[last_match:i].split()
        [tokens.append(t) for t in match if t]

    @classmethod
    @override
    def tokenize(cls, s: str) -> List[str]:
        if not s:
            return []

        tokens: List[str] = []

        i = 0  # ensure i is bound
        last_match = 0
        last_membership = s[0] in cls.delimiters
        for i, char in enumerate(s):
            mem = char in cls.delimiters
            ucsur = NimiUCSUR.filter(char)  # always "changed" means
            changed = (mem != last_membership) or ucsur
            # this keeps contiguous words together, but splits UCSUR
            if not changed:
                continue

            if ucsur:
                if i > last_match:
                    # Add the token before UCSUR character
                    cls.__helper(s, tokens, last_match, i)
                # Add UCSUR character itself as a token
                tokens.append(char)
                last_match = i + 1
                last_membership = mem
                continue

            cls.__helper(s, tokens, last_match, i)
            last_match = i
            last_membership = mem

        cls.__helper(s, tokens, last_match, i + 1)
        return tokens


@deprecated(
    "WordTokenizerRe is a previous reference implementation. Its behavior has diverged from WordTokenizer and it may not be restored."
)
class WordTokenizerRe(RegexTokenizer):
    pattern = re.compile(rf"""([{ALL_PUNCT_RANGES}]+|\s+)""")


@deprecated(
    "WordTokenizerRe1 is a previous reference implementation. Its behavior has diverged from WordTokenizer and it may not be restored."
)
class WordTokenizerRe1(Regex1Tokenizer):
    """Reference implementation for WordTokenizer."""

    pattern = regex.compile(r"""([\p{posix_punct}\p{Punctuation}]+|\s+)""")


class SentTokenizer(SetTokenizer):
    delimiters = set(SENTENCE_PUNCT + "\n")  # regex does \n with a flag

    @classmethod
    @override
    def tokenize(cls, s: str) -> List[str]:
        if not s:
            return []

        tokens: List[str] = []
        last_match = 0
        for i, char in enumerate(s):
            if char not in cls.delimiters:
                continue

            match = s[last_match : i + 1].strip()
            last_match = i + 1  # newlines can strip but idc
            if not match:
                continue
            tokens.append(match)

        match = s[last_match:].strip()
        if match:
            tokens.append(match)

        return tokens


class SentTokenizerRe(RegexTokenizer):
    pattern = re.compile(
        rf"""(?<=[{regex_escape(SENTENCE_PUNCT)}])|$""", flags=re.MULTILINE
    )
    # TODO: are <> or {} that common as *sentence* delims? [] are already a stretch
    # TODO: do the typography characters matter?
    # NOTE: | / and , are *not* sentence delimiters for my purpose


class SentTokenizerRe1(Regex1Tokenizer):
    pattern = regex.compile(
        rf"""(?<=[{regex_escape(SENTENCE_PUNCT)}]|$)""", flags=regex.MULTILINE
    )


__all__ = [
    "WordTokenizer",
    "SentTokenizer",
]



# STL
from typing import List, Type, Tuple

# LOCAL
from sonatoki.Filters import Filter
from sonatoki.Scorers import Number, Scorer
from sonatoki.Cleaners import Cleaner
from sonatoki.Tokenizers import Tokenizer, SentTokenizer, WordTokenizer
from sonatoki.Preprocessors import Preprocessor

# tokenized, filtered, cleaned, score, result
Scorecard = Tuple[List[str], List[str], List[str], Number, bool]
# TODO: scorecard kinda sucks as a name


class Ilo:
    __preprocessors: List[Type[Preprocessor]]
    __sent_tokenizer: Type[Tokenizer]
    __word_tokenizer: Type[Tokenizer]
    __cleaners: List[Type[Cleaner]]
    __ignoring_filters: List[Type[Filter]]
    __scoring_filters: List[Type[Filter]]
    __scorer: Type[Scorer]
    __passing_score: Number

    def __init__(
        self,
        preprocessors: List[Type[Preprocessor]],
        cleaners: List[Type[Cleaner]],
        ignoring_filters: List[Type[Filter]],
        scoring_filters: List[Type[Filter]],
        scorer: Type[Scorer],
        passing_score: Number,
        word_tokenizer: Type[Tokenizer] = WordTokenizer,
        sent_tokenizer: Type[Tokenizer] = SentTokenizer,
    ):
        super().__init__()
        # avoid keeping a ref to user's list just in case
        self.__preprocessors = [*preprocessors]
        self.__sent_tokenizer = sent_tokenizer
        self.__word_tokenizer = word_tokenizer
        self.__cleaners = [*cleaners]
        self.__ignoring_filters = [*ignoring_filters]
        self.__scoring_filters = [*scoring_filters]
        self.__scorer = scorer
        self.__passing_score = passing_score

    def preprocess(self, msg: str) -> str:
        for p in self.__preprocessors:
            msg = p.process(msg)
        return msg

    def word_tokenize(self, msg: str) -> List[str]:
        """It is *highly* recommended that you run `ilo.preprocess` first."""
        return self.__word_tokenizer.tokenize(msg)

    def sent_tokenize(self, msg: str) -> List[str]:
        return self.__sent_tokenizer.tokenize(msg)

    def clean_token(self, token: str) -> str:
        for c in self.__cleaners:
            token = c.clean(token)
        return token

    def clean_tokens(self, tokens: List[str]) -> List[str]:
        # NOTE: tested, making a new list with a for loop *is* faster than:
        # list comp, generator comp, in-place replacement
        cleaned_tokens: List[str] = list()
        for token in tokens:
            cleaned_token = self.clean_token(token)
            if not cleaned_token:
                # TODO: warn user?
                continue
            cleaned_tokens.append(cleaned_token)
        return cleaned_tokens

    def _filter_token(self, token: str) -> bool:
        for f in self.__ignoring_filters:
            if f.filter(token):
                return True
        return False

    def filter_tokens(self, tokens: List[str]) -> List[str]:
        filtered_tokens: List[str] = []
        for token in tokens:
            if self._filter_token(token):
                continue
            # the ignoring filter is true if the token matches
            # the user wants to ignore these so keep non-matching tokens
            filtered_tokens.append(token)
        return filtered_tokens

    def score_tokens(self, tokens: List[str]) -> float:
        return self.__scorer.score(tokens, self.__scoring_filters)

    def _is_toki_pona(self, message: str) -> Scorecard:
        """Process a message into its tokens, then filters, cleans, and scores
        them. Returns all parts. Message must already be preprocessed, normally
        done in `self.is_toki_pona(message)`.

        Returns all components of the processing algorithm except preprocessing:
        - Tokenized message (list[str])
        - Filtered message (list[str])
        - Cleaned message (list[str])
        - Score (float)
        - Result (bool)
        """
        tokenized = self.word_tokenize(message)
        filtered = self.filter_tokens(tokenized)
        cleaned = self.clean_tokens(filtered)
        score = self.score_tokens(cleaned)
        result = score >= self.__passing_score

        return tokenized, filtered, cleaned, score, result

    def is_toki_pona(self, message: str) -> bool:
        """Determines whether a single statement is or is not Toki Pona."""
        message = self.preprocess(message)
        *_, result = self._is_toki_pona(message)
        return result

    def _are_toki_pona(self, message: str):
        """Split a message into sentences, then return a list each sentence's
        results via `self._is_toki_pona()`.

        Message must already be preprocessed, normally done in
        `self.are_toki_pona(message)`.
        """
        results: List[Scorecard] = list()
        for sentence in self.sent_tokenize(message):
            result = self._is_toki_pona(sentence)
            results.append(result)
        return results

    def are_toki_pona(self, message: str) -> List[bool]:
        """Splits a statement into sentences, then determines if each is or is not Toki Pona.
        NOTE: You will need to decide how to score the result. Examples:

        ```
        def all_must_pass(message: str) -> bool:
            return all(ILO.are_toki_pona(message))

        def portion_must_pass(message: str, score: Number = 0.8) -> bool:
            results = ILO.are_toki_pona(message)
            sent_count = len(results)
            passing = results.count(True)
            return (passing / sent_count) >= score
        ```
        """
        message = self.preprocess(message)
        results = self._are_toki_pona(message)
        return [res[-1] for res in results]
